@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2015 Canonical Ltd
+ * Copyright (C) 2015 Canonical Ltd.
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License version 3 as
@@ -40,7 +40,13 @@ PersistentStringCache provides a cache of key-value pairs with a backing store.
 It is intended for caching arbitrary (possibly large) amounts of data,
 such as might be needed by a web browser cache. The cache scales to large
 numbers (hundreds of thousands) of entries and is very fast. (Typically,
-the performance limiting factor is the I/O bandwidth to disk.)
+the performance-limiting factor is the I/O bandwidth to disk.)
+
+The cache is robust in the face of crashes and power loss. After a
+re-start, it is guaranteed to be in a consistent state with correct
+data. Some number of updates that were made just prior to a power loss
+or kernel crash can be lost; however, if just the calling process
+crashes, no updates are lost.
 
 A cache has a maximum size (which can be changed at any time). Once
 the cache reaches its maximum size, when adding an entry,
@@ -181,9 +187,6 @@ public:
 
     /**
     Destroys the instance.
-
-    The destructor compacts the database. This ensures that, while a cache is
-    not in use, it comsumes as little disk space as possible.
     */
     ~PersistentStringCache();
 
@@ -307,7 +310,12 @@ public:
 
     /**
     \brief Returns statistics for the cache.
+
+    The returned statistics are persistent and are restored the next
+    time an existing cache is opened. Call clear_stats() to explicitly
+    reset the statistics counters and time stamps to zero.
     \return An object that provides accessors to statistics and settings.
+    \see clear_stats()
     */
     PersistentCacheStats stats() const;
 
@@ -586,7 +594,10 @@ public:
 
     /**
     \brief Deletes all entries from the cache.
-    \note This operation compacts the database to use the smallest possible amount of disk space.
+
+    This operation completely empties the cache.
+    \note Clearing the cache also resets the statistics counters.
+    \see clear_stats()
     */
     void invalidate();
 
@@ -622,7 +633,8 @@ public:
 
     \throws invalid_argument `size_in_bytes` is &lt; 1
 
-    \note This operation compacts the database to use the smallest possible amount of disk space.
+    \note If the new size is less than the current size, this operation compacts the database
+    to use the smallest possible amount of disk space.
     */
     void resize(int64_t size_in_bytes);
 
@@ -635,12 +647,18 @@ public:
 
     \throws invalid_argument `used_size_in_bytes` is &lt; 0
     \throws logic_error `used_size_in_bytes` is &gt; max_size_in_bytes().
-
-    \note If trimming actually took place, this operation compacts the database to use the
-    smallest possible amount of disk space.
     */
     void trim_to(int64_t used_size_in_bytes);
 
+    /**
+    \brief Compacts the database.
+
+    This operation compacts the database to consume as little disk space as possible.
+    Note that this operation can be slow. (Compacting a 100 MB cache can take around
+    ten seconds on a machine with a spinning-platter disk.)
+    */
+    void compact();
+
     //@}
 
     /** @name Monitoring cache activity
